"""Constants used by ortografix."""

import torch

__all__ = ('EOS', 'SOS', 'SEP', 'DEVICE', 'SOS_IDX', 'EOS_IDX')

DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
EOS = '__EOS__'
SOS = '__SOS__'
SEP = '__SEP__'
SOS_IDX = 0
EOS_IDX = 1



"""Dataset model."""
import os
import random
import logging

import ortografix.utils.processing as putils

__all__ = ('Dataset', 'Vocab')

logger = logging.getLogger(__name__)


class Vocab():
    """A Vocab class to process vocabularies for source/target sequences."""

    def __init__(self, dataset_filepath=None, vocab_filepath=None,
                 is_character_based=False, is_source=True, is_reversed=False):
        """Initialize vocabulary."""
        if not (dataset_filepath or vocab_filepath):
            raise Exception('You must specify either dataset_filepath or '
                            'vocab_filepath to initialize the vocabulary')
        if dataset_filepath and vocab_filepath:
            raise Exception('You cannot specify *both* dataset_filepath AND '
                            'vocab_filepath to initialize the vocabulary')
        if dataset_filepath:
            self._vocab = putils.create_vocab(
                dataset_filepath, is_character_based, is_source, is_reversed)
        if vocab_filepath:
            self._vocab = putils.load_vocab(vocab_filepath)

    @property
    def size(self):
        """Return the number of items in the vocabulary."""
        return len(self._vocab)

    @property
    def item2idx(self):
        """Return token/character-to-idx mapping (dict)."""
        return self._vocab

    @property
    def idx2item(self):
        """Return idx-to-token/character mapping (dict)."""
        return {idx: item for item, idx in self._vocab.items()}


# pylint: disable=R0902
class Dataset():
    """A dataset class to return source/target tensors from training data."""

    def __init__(self, data_filepath, is_character_based, shuffle,
                 max_seq_len, is_reversed):
        """Prepare input tensors.

        Prepare dictionaries for source and target items.
        Discretize input to indexes.
        Convert to tensors and concatenate by batch
        """
        self._data_filepath = data_filepath
        self._is_character_based = is_character_based
        self._shuffle = shuffle
        self._is_reversed = is_reversed
        self._max_seq_len = max_seq_len
        self._source_vocab = Vocab(dataset_filepath=data_filepath,
                                   is_character_based=is_character_based,
                                   is_source=True, is_reversed=is_reversed)
        self._target_vocab = Vocab(dataset_filepath=data_filepath,
                                   is_character_based=is_character_based,
                                   is_source=False, is_reversed=is_reversed)
        self._indexes = putils.index_dataset(
            self._data_filepath, self._source_vocab.item2idx,
            self._target_vocab.item2idx, self._is_character_based,
            self._max_seq_len, self._is_reversed)

    @property
    def is_character_based(self):
        """Return True if character-based, False if token-based."""
        return self._is_character_based

    @property
    def shuffle(self):
        """Return True if dataset sentence pairs will be shuffled."""
        return self._shuffle

    @property
    def max_seq_len(self):
        """Return the maximal sequence length to be considered."""
        return self._max_seq_len

    @property
    def source_vocab(self):
        """Return vocabulary of source."""
        return self._source_vocab

    @property
    def target_vocab(self):
        """Return vocabulary of target."""
        return self._target_vocab

    @property
    def indexes(self):
        """Return a list of source/target indexed sequences."""
        if self._shuffle:
            random.shuffle(self._indexes)
        return self._indexes

    def save_params(self, output_dirpath):
        """Save vocabularies and dataset parameters."""
        logger.info('Saving dataset to directory {}'.format(output_dirpath))
        params_filepath = os.path.join(output_dirpath, 'dataset.params')
        with open(params_filepath, 'w', encoding='utf-8') as output_str:
            print('shuffle\t{}'.format(self._shuffle), file=output_str)
            print('is_character_based\t{}'.format(self._is_character_based),
                  file=output_str)
            print('max_seq_len\t{}'.format(self._max_seq_len), file=output_str)
            print('is_reversed\t{}'.format(self._is_reversed), file=output_str)
        logger.info('Saving source vocab...')
        source_vocab_filepath = os.path.join(output_dirpath, 'source.vocab')
        with open(source_vocab_filepath, 'w', encoding='utf-8') as source_str:
            for item, idx in self._source_vocab.item2idx.items():
                print('{}\t{}'.format(item, idx), file=source_str)
        logger.info('Saving target vocab...')
        target_vocab_filepath = os.path.join(output_dirpath, 'target.vocab')
        with open(target_vocab_filepath, 'w', encoding='utf-8') as target_str:
            for item, idx in self._target_vocab.item2idx.items():
                print('{}\t{}'.format(item, idx), file=target_str)



"""Process data and params for training/decoding."""
import sys
import logging

import ortografix.utils.constants as const

logger = logging.getLogger(__name__)

__all__ = ('index_dataset', 'index_sequence', 'create_vocab', 'load_vocab',
           'load_params', 'get_max_seq_len')


def get_max_seq_len(data_filepath, is_character_based):
    """Return max sequence length computed from dataset.

    If is_character_based, will return the length of the longuest word.
    Else, will return the length of the longuest sentence.
    """
    if is_character_based:
        if not _is_dataset_consistent(data_filepath):
            logger.error('Dataset in inconsistent state. Exiting...')
            sys.exit(0)
    max_seq_len = 0
    with open(data_filepath, 'r', encoding='utf-8') as data_str:
        for line in data_str:
            line = line.strip()
            tokens = line.split('\t')[0].split()
            if is_character_based:
                for token in tokens:
                    max_seq_len = max(max_seq_len, len(token))
            else:
                max_seq_len = max(max_seq_len, len(tokens))
            tokens = line.split('\t')[1].split()
            if is_character_based:
                for token in tokens:
                    max_seq_len = max(max_seq_len, len(token))
            else:
                max_seq_len = max(max_seq_len, len(tokens))
    return max_seq_len


def load_params(params_filepath):
    """Load parameter dict from file (tab separated)."""
    params = {}
    with open(params_filepath, 'r', encoding='utf-8') as params_str:
        for line in params_str:
            line = line.strip()
            items = line.split('\t')
            if items[0] in ['shuffle', 'is_character_based', 'cuda',
                            'is_reversed']:
                params[items[0]] = items[1] == 'True'
            elif items[0] in ['max_seq_len']:
                params[items[0]] = int(items[1])
            else:
                raise Exception('Unsupported dataset parameter: {}'
                                .format(items[0]))
    return params


def load_vocab(vocab_filepath):
    """Load previously saved vocab to dic."""
    vocab = {}
    with open(vocab_filepath, 'r', encoding='utf-8') as vocab_str:
        for line in vocab_str:
            line = line.strip()
            items = line.split('\t')
            vocab[items[0]] = int(items[1])

    return vocab


def _create_vocab(input_str, is_character_based, is_source, is_reversed):
    logger.info('Preparing source and target dictionaries...')
    vocab = {const.SOS: const.SOS_IDX, const.EOS: const.EOS_IDX}
    ref_idx = 1 if is_source == is_reversed else 0
    for line in input_str:
        line = line.strip()
        sent = line.split('\t')[ref_idx]
        for token in sent.split():
            if is_character_based is True:
                for char in token:
                    if char not in vocab:
                        vocab[char] = len(vocab)
            elif is_character_based is False:
                if token not in vocab:
                    vocab[token] = len(vocab)
            else:
                raise Exception('Unsupported value {} for is_character_based '
                                'parameter'.format(is_character_based))
    desc = 'Source' if is_source else 'Target'
    logger.info('{} vocabulary contains {} items'.format(desc, len(vocab)))
    return vocab


def create_vocab(dataset_filepath, character_based, is_source, is_reversed):
    """Generate token/character-to-idx mapping for a dataset file.

    Dataset file should contains pairs of aligned sentences,
    even in character-based mode.
    source_or_target should specify whether vocab should be created for source
    or target sequence
    """
    with open(dataset_filepath, 'r', encoding='utf-8') as input_str:
        return _create_vocab(input_str, character_based, is_source,
                             is_reversed)


def index_sequence(sequence, item2idx, is_character_based):
    """Convert a sequence to a list of int following item2idx.

    TODO: add support for OOV.
    """
    indexes = [const.SOS]
    sequence = sequence.strip()
    for item in sequence.split():
        if is_character_based:
            raise Exception('Unsupported is_character_based option for now')
        indexes.append(item2idx[item])
    indexes.append(const.EOS)
    return indexes


def _index_tokens(input_stream, source_item2idx, target_item2idx, max_seq_len,
                  is_reversed):
    logger.info('Preparing token-based source-target indexes...')
    source_target_indexes = []
    for line in input_stream:
        line = line.strip()
        source_sent = line.split('\t')[1] if is_reversed \
         else line.split('\t')[0]
        target_sent = line.split('\t')[0] if is_reversed \
         else line.split('\t')[1]
        source_tokens = source_sent.split()
        target_tokens = target_sent.split()
        if len(source_tokens) > max_seq_len \
         or len(target_tokens) > max_seq_len:
            continue
        source_tokens = [const.SOS] + source_tokens  # prepend value
        source_tokens.append(const.EOS)
        target_tokens = [const.SOS] + target_tokens  # prepend value
        target_tokens.append(const.EOS)
        source_indexes = [source_item2idx[token] for token in source_tokens]
        target_indexes = [target_item2idx[token] for token in target_tokens]
        source_target_indexes.append((source_indexes, target_indexes))
    logger.info('Source-target indexes contain {} pairs of items'
                .format(len(source_target_indexes)))
    return source_target_indexes


# pylint: disable=R0914
def _index_characters(input_stream, source_item2idx, target_item2idx,
                      max_seq_len, is_reversed):
    logger.info('Preparing character-based source-target indexes...')
    source_target_indexes = []
    for line in input_stream:
        line = line.strip()
        source_sent = line.split('\t')[1] if is_reversed \
         else line.split('\t')[0]
        target_sent = line.split('\t')[0] if is_reversed \
         else line.split('\t')[1]
        source_tokens = source_sent.split()
        target_tokens = target_sent.split()
        for source_token, target_token in zip(source_tokens,
                                              target_tokens):
            source_indexes = []
            target_indexes = []
            if len(source_token) > max_seq_len \
             or len(target_token) > max_seq_len:
                # careful! This can remove words in the middle of a sentence
                continue
            # Here, string sequence is taken to be word/token. SOS indicates
            # the start of a token, EOS its end.
            source_indexes.append(source_item2idx[const.SOS])
            target_indexes.append(target_item2idx[const.SOS])
            for source_char in source_token:
                source_indexes.append(source_item2idx[source_char])
            for target_char in target_token:
                target_indexes.append(target_item2idx[target_char])
            source_indexes.append(source_item2idx[const.EOS])
            target_indexes.append(target_item2idx[const.EOS])
            source_target_indexes.append((source_indexes, target_indexes))
    logger.info('Source-target indexes contain {} pairs of items'
                .format(len(source_target_indexes)))
    return source_target_indexes


def _is_dataset_consistent(data_filepath):
    """Check that all lines contains the same number of tokens.

    This is necessary only for character-based indexing.
    """
    is_consistent = True
    with open(data_filepath, 'r', encoding='utf-8') as data_str:
        for line in data_str:
            line = line.strip()
            source_sent = line.split('\t')[0]
            target_sent = line.split('\t')[1]
            source_tokens = source_sent.split()
            target_tokens = target_sent.split()
            if len(source_tokens) != len(target_tokens):
                is_consistent = False
                logger.error('Source and target sequences do NOT contain same '
                             'number of tokens in: \n source: {}\n target: {}'
                             .format(source_sent, target_sent))
    return is_consistent


def index_dataset(data_filepath, source_item2idx, target_item2idx,
                  is_character_based, max_seq_len, is_reversed):
    """Convert input data text file to list of indexed integers."""
    logger.info('Indexing dataset...')
    if is_character_based:
        if not _is_dataset_consistent(data_filepath):
            logger.error('Dataset in inconsistent state. Exiting...')
            sys.exit(0)
    with open(data_filepath, 'r', encoding='utf-8') as data_str:
        if is_character_based:
            return _index_characters(data_str, source_item2idx,
                                     target_item2idx, max_seq_len, is_reversed)
        return _index_tokens(data_str, source_item2idx, target_item2idx,
                             max_seq_len, is_reversed)



"""Attention Decoder."""
import torch

import ortografix.utils.constants as const

__all__ = ('Attention')


# pylint: disable=R0902
class Attention(torch.nn.Module):
    """Attention class."""

    def __init__(self, model_type, hidden_size, output_size, max_seq_len,
                 num_layers=1, nonlinearity='tanh', bias=True, dropout=0,
                 bidirectional=False):
        """Initialize attention model."""
        super(Attention, self).__init__()
        self.model_type = model_type
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.max_seq_len = max_seq_len+2  # add SOS and EOS
        self.num_layers = num_layers
        self.nonlinearity = nonlinearity
        self.bias = bias
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.embedding = torch.nn.Embedding(self.output_size, self.hidden_size)
        self.attn = torch.nn.Linear(self.hidden_size * 2, self.max_seq_len)
        if self.bidirectional:
            self.attn_combine = torch.nn.Linear(self.hidden_size * 3,
                                                self.hidden_size)
        else:
            self.attn_combine = torch.nn.Linear(self.hidden_size * 2,
                                                self.hidden_size)
        if self.model_type == 'rnn':
            self.rnn = torch.nn.RNN(
                input_size=hidden_size, hidden_size=hidden_size,
                num_layers=num_layers, nonlinearity=nonlinearity,
                bias=bias, batch_first=False, dropout=dropout,
                bidirectional=bidirectional)
        if self.model_type == 'gru':
            self.gru = torch.nn.GRU(
                input_size=hidden_size, hidden_size=hidden_size,
                num_layers=num_layers, bias=bias, batch_first=False,
                dropout=dropout, bidirectional=bidirectional)
        if self.model_type == 'lstm':
            self.lstm = torch.nn.LSTM(
                input_size=hidden_size, hidden_size=hidden_size,
                num_layers=num_layers, bias=bias, batch_first=False,
                dropout=dropout, bidirectional=bidirectional)
        if self.bidirectional:
            self.out = torch.nn.Linear(self.hidden_size*2, self.output_size)
        else:
            self.out = torch.nn.Linear(self.hidden_size, self.output_size)

    # pylint: disable=R1710, W0221
    def forward(self, input_tensor, hidden, encoder_outputs):
        """Apply forward computation."""
        embedded = self.embedding(input_tensor).view(1, 1, -1)
        attn_weights = torch.nn.functional.softmax(
            self.attn(torch.cat((embedded[0], hidden[0]), 1)), dim=1)
        attn_applied = torch.bmm(attn_weights.unsqueeze(0),
                                 encoder_outputs.unsqueeze(0))
        output = torch.cat((embedded[0], attn_applied[0]), 1)
        output = self.attn_combine(output).unsqueeze(0)
        output = torch.nn.functional.relu(output)
        if self.model_type == 'rnn':
            output, hidden = self.rnn(output, hidden)
        elif self.model_type == 'gru':
            output, hidden = self.gru(output, hidden)
        else:
            output, hidden = self.lstm(output, hidden)
        output = torch.nn.functional.log_softmax(self.out(output[0]), dim=1)
        return output, hidden, attn_weights

    def init_hidden(self):
        """Initialize hidden layers."""
        return torch.zeros(self.num_layers, 1, self.hidden_size,
                           device=const.DEVICE)



"""Time utils."""
import time
import math

__all__ = ('time_since')


def as_minutes(seconds):
    """Convert input to minutes."""
    minutes = math.floor(seconds / 60)
    seconds -= minutes * 60
    return '{}m {}s'.format(minutes, round(seconds))


def time_since(since, percent):
    """Return time since."""
    now = time.time()
    elapsed = now - since
    ratio = elapsed / percent
    remaining = ratio - elapsed
    return '{} (- {})'.format(as_minutes(elapsed), as_minutes(remaining))



"""Welcome to ortografix.

This is the entry point of the application.
"""
import os

import argparse
import random
import time
import statistics
import logging
import logging.config

import textdistance
import torch
from torch import optim

import ortografix.utils.config as cutils
import ortografix.utils.constants as const
import ortografix.utils.time as tutils
import ortografix.utils.processing as putils

from ortografix.model.attention import Attention
from ortografix.model.encoder import Encoder
from ortografix.model.decoder import Decoder
from ortografix.model.dataset import Dataset, Vocab


logging.config.dictConfig(
    cutils.load(
        os.path.join(os.path.dirname(__file__), 'logging', 'logging.yml')))

logger = logging.getLogger(__name__)


__all__ = ('train', 'evaluate')


def save_dataset_and_models(output_dirpath, dataset, encoder, decoder, loss,
                            learning_rate, with_attention):
    """Dump pytorch model and Dataset parameters."""
    logger.info('Saving dataset and models...')
    dataset.save_params(output_dirpath)
    model_params_filepath = os.path.join(output_dirpath, 'model.params')
    with open(model_params_filepath, 'w', encoding='utf-8') as m_params_str:
        print('cuda\t{}'.format(torch.cuda.is_available()), file=m_params_str)
    torch.save({'encoder_state_dict': encoder.state_dict(),
                'decoder_state_dict': decoder.state_dict(),
                'encoder': {
                    'model_type': encoder.model_type,
                    'input_size': encoder.input_size,
                    'hidden_size': encoder.hidden_size,
                    'num_layers': encoder.num_layers,
                    'nonlinearity': encoder.nonlinearity,
                    'bias': encoder.bias,
                    'dropout': encoder.dropout,
                    'bidirectional': encoder.bidirectional
                },
                'decoder': {
                    'model_type': decoder.model_type,
                    'output_size': decoder.output_size,
                    'hidden_size': decoder.hidden_size,
                    'num_layers': decoder.num_layers,
                    'nonlinearity': decoder.nonlinearity,
                    'bias': decoder.bias,
                    'dropout': decoder.dropout,
                    'bidirectional': decoder.bidirectional
                },
                'with_attention': with_attention,
                'loss': loss,
                'learning_rate': learning_rate},
               os.path.join(output_dirpath, 'checkpoint.tar'))


# pylint: disable=R0914,E1102
def _train_single_batch(source_tensor, target_tensor, encoder, decoder,
                        with_attention, encoder_optimizer, decoder_optimizer,
                        max_seq_len, criterion, use_teacher_forcing,
                        teacher_forcing_ratio):
    encoder_hidden = encoder.init_hidden()
    encoder_optimizer.zero_grad()
    decoder_optimizer.zero_grad()
    input_length = source_tensor.size(0)
    target_length = target_tensor.size(0)
    # add 2 to max_seq_len to include SOS and EOS
    if encoder.bidirectional:
        encoder_outputs = torch.zeros(max_seq_len+2, encoder.hidden_size*2,
                                      device=const.DEVICE)
    else:
        encoder_outputs = torch.zeros(max_seq_len+2, encoder.hidden_size,
                                      device=const.DEVICE)
    loss = 0
    for eidx in range(input_length):
        encoder_output, encoder_hidden = encoder(source_tensor[eidx],
                                                 encoder_hidden)
        encoder_outputs[eidx] = encoder_output[0, 0]
    decoder_input = torch.tensor([[const.SOS_IDX]], device=const.DEVICE)
    decoder_hidden = encoder_hidden
    use_teacher_forcing = random.random() < teacher_forcing_ratio
    if use_teacher_forcing:
        # Teacher forcing: Feed the target as the next input
        for didx in range(target_length):
            if with_attention:
                decoder_output, decoder_hidden, _ = decoder(
                    decoder_input, decoder_hidden, encoder_outputs)
            else:
                decoder_output, decoder_hidden = decoder(
                    decoder_input, decoder_hidden)
            loss += criterion(decoder_output, target_tensor[didx])
            decoder_input = target_tensor[didx]  # Teacher forcing
    else:
        # Without teacher forcing: use its own predictions as the next input
        for didx in range(target_length):
            if with_attention:
                decoder_output, decoder_hidden, _ = decoder(
                    decoder_input, decoder_hidden, encoder_outputs)
            else:
                decoder_output, decoder_hidden = decoder(
                    decoder_input, decoder_hidden)
            _, topi = decoder_output.topk(1)
            # detach from history as input
            decoder_input = topi.squeeze().detach()
            loss += criterion(decoder_output, target_tensor[didx])
            if decoder_input.item() == const.EOS_IDX:
                break
    loss.backward()
    encoder_optimizer.step()
    decoder_optimizer.step()
    return loss.item() / target_length


def _train(encoder, decoder, dataset, with_attention, num_epochs,
           learning_rate, print_every, use_teacher_forcing,
           teacher_forcing_ratio, output_dirpath):
    encoder_optimizer = optim.SGD(encoder.parameters(), lr=learning_rate)
    decoder_optimizer = optim.SGD(decoder.parameters(), lr=learning_rate)
    criterion = torch.nn.NLLLoss()
    start = time.time()
    print_loss_total = 0  # Reset every print_every
    num_iter = 0
    num_total_iters = len(dataset.indexes) * num_epochs
    try:
        logger.info('Starting training...')
        for epoch in range(1, num_epochs+1):
            for source_indexes, target_indexes in dataset.indexes:
                source_tensor = torch.tensor(source_indexes, dtype=torch.long,
                                             device=const.DEVICE).view(-1, 1)
                target_tensor = torch.tensor(target_indexes, dtype=torch.long,
                                             device=const.DEVICE).view(-1, 1)
                num_iter += 1
                loss = _train_single_batch(
                    source_tensor, target_tensor, encoder, decoder,
                    with_attention, encoder_optimizer, decoder_optimizer,
                    dataset.max_seq_len, criterion, use_teacher_forcing,
                    teacher_forcing_ratio)
                print_loss_total += loss
                if num_iter % print_every == 0:
                    print_loss_avg = print_loss_total / print_every
                    print_loss_total = 0
                    time_info = tutils.time_since(start,
                                                  num_iter / num_total_iters)
                    logger.info('Epoch {}/{} {} ({} {}%) {}'.format(
                        epoch, num_epochs, time_info, num_iter,
                        round(num_iter / num_total_iters * 100),
                        round(print_loss_avg, 4)))
        save_dataset_and_models(output_dirpath, dataset, encoder, decoder,
                                loss, learning_rate, with_attention)
    except KeyboardInterrupt:
        logger.info('Training interrupted')
        save_dataset_and_models(output_dirpath, dataset, encoder, decoder,
                                loss, learning_rate, with_attention)


def train(args):
    """Train the model."""
    logger.info('Training model from {}'.format(args.data))
    if not os.path.exists(args.output_dirpath):
        logger.info('Creating output directory to save files to: {}'
                    .format(args.output_dirpath))
        os.makedirs(args.output_dirpath, exist_ok=True)
    if torch.cuda.is_available():
        logger.info('Training on GPU')
    else:
        logger.info('No GPU available. Training on CPU')
    if not args.max_seq_len:
        max_seq_len = putils.get_max_seq_len(args.data, args.character_based)
        logger.info('max_seq_len not specified in args. Automatically setting '
                    'to longest source sequence length = {}'
                    .format(max_seq_len))
    else:
        logger.info('Manually setting max_seq_len to {}'
                    .format(args.max_seq_len))
        max_seq_len = args.max_seq_len
    dataset = Dataset(args.data, args.character_based, args.shuffle,
                      max_seq_len, args.reverse)
    encoder = Encoder(model_type=args.model_type,
                      input_size=dataset.source_vocab.size,
                      hidden_size=args.hidden_size,
                      num_layers=args.num_layers,
                      nonlinearity=args.nonlinearity,
                      bias=args.bias, dropout=args.dropout,
                      bidirectional=args.bidirectional).to(const.DEVICE)
    if args.with_attention:
        decoder = Attention(model_type=args.model_type,
                            hidden_size=args.hidden_size,
                            output_size=dataset.target_vocab.size,
                            max_seq_len=dataset.max_seq_len,
                            num_layers=args.num_layers,
                            nonlinearity=args.nonlinearity,
                            bias=args.bias, dropout=args.dropout,
                            bidirectional=args.bidirectional).to(const.DEVICE)
    else:
        decoder = Decoder(model_type=args.model_type,
                          hidden_size=args.hidden_size,
                          output_size=dataset.target_vocab.size,
                          num_layers=args.num_layers,
                          nonlinearity=args.nonlinearity,
                          bias=args.bias, dropout=args.dropout,
                          bidirectional=args.bidirectional).to(const.DEVICE)
    return _train(encoder, decoder, dataset, args.with_attention, args.epochs,
                  args.learning_rate, args.print_every,
                  args.use_teacher_forcing, args.teacher_forcing_ratio,
                  args.output_dirpath)


def _decode(source_indexes, encoder, decoder, with_attention, max_seq_len):
    with torch.no_grad():
        input_tensor = torch.tensor(source_indexes, dtype=torch.long,
                                    device=const.DEVICE).view(-1, 1)
        input_length = input_tensor.size()[0]
        encoder_hidden = encoder.init_hidden()
        # add 2 to max_seq_len to include SOS and EOS
        if encoder.bidirectional:
            encoder_outputs = torch.zeros(max_seq_len+2, encoder.hidden_size*2,
                                          device=const.DEVICE)
        else:
            encoder_outputs = torch.zeros(max_seq_len+2, encoder.hidden_size,
                                          device=const.DEVICE)
        for eidx in range(input_length):
            encoder_output, encoder_hidden = encoder(
                input_tensor[eidx], encoder_hidden)
            encoder_outputs[eidx] += encoder_output[0, 0]
        decoder_input = torch.tensor([[const.SOS_IDX]], device=const.DEVICE)
        decoder_hidden = encoder_hidden
        decoded_indexes = []
        # add 2 to max_seq_len to include SOS and EOS
        decoder_attentions = torch.zeros(max_seq_len+2, max_seq_len+2)
        for didx in range(max_seq_len+2):
            if with_attention:
                decoder_output, decoder_hidden, decoder_attention = decoder(
                    decoder_input, decoder_hidden, encoder_outputs)
                decoder_attentions[didx] = decoder_attention.data
            else:
                decoder_output, decoder_hidden = decoder(decoder_input,
                                                         decoder_hidden)
            _, topi = decoder_output.data.topk(1)
            if topi.item() == const.EOS_IDX:
                decoded_indexes.append(const.EOS_IDX)
                break
            decoded_indexes.append(topi.item())
            decoder_input = topi.squeeze().detach()
        return decoded_indexes, decoder_attentions[:didx + 1]


def _evaluate(indexes, encoder, decoder, target_vocab, checkpoint,
              dataset_params):
    avg_dist = []
    avg_norm_dist = []
    avg_norm_sim = []
    for seq in indexes:
        pred_idx, _ = _decode(seq[0], encoder, decoder,
                              checkpoint['with_attention'],
                              dataset_params['max_seq_len'])
        gold = ''.join(
            [target_vocab.idx2item[idx] for idx in seq[1] if idx not in
             [const.SOS_IDX, const.EOS_IDX]])
        prediction = ''.join(
            [target_vocab.idx2item[idx] for idx in pred_idx if idx not in
             [const.SOS_IDX, const.EOS_IDX]])
        avg_dist.append(textdistance.levenshtein.distance(gold,
                                                          prediction))
        avg_norm_dist.append(textdistance.levenshtein.normalized_distance(
            gold, prediction))
        avg_norm_sim.append(textdistance.levenshtein.normalized_similarity(
            gold, prediction))
    logger.info('Printing Levenshtein edit distance info:')
    logger.info('   total sum dist = {}'.format(sum(avg_dist)))
    logger.info('   avg dist = {}'.format(statistics.mean(avg_dist)))
    # logger.info('   avg sim = {}'.format(statistics.mean(avg_sim)))
    logger.info('   avg normalized dist = {}'
                .format(statistics.mean(avg_norm_dist)))
    logger.info('   avg normalized sim = {}'
                .format(statistics.mean(avg_norm_sim)))
    return avg_dist, avg_norm_dist, avg_norm_sim


def evaluate(args):
    """Evaluate a given model on a test set."""
    dataset_param_filepath = os.path.join(args.model, 'dataset.params')
    dataset_params = putils.load_params(dataset_param_filepath)
    source_vocab_filepath = os.path.join(args.model, 'source.vocab')
    source_vocab = Vocab(vocab_filepath=source_vocab_filepath)
    target_vocab_filepath = os.path.join(args.model, 'target.vocab')
    target_vocab = Vocab(vocab_filepath=target_vocab_filepath)
    model_params_filepath = os.path.join(args.model, 'model.params')
    model_params = putils.load_params(model_params_filepath)
    checkpoint_filepath = os.path.join(args.model, 'checkpoint.tar')
    if not torch.cuda.is_available() and model_params['cuda']:
        logger.info('Loading a GPU-trained model on CPU')
        checkpoint = torch.load(checkpoint_filepath,
                                map_location=const.DEVICE)
    elif torch.cuda.is_available() and model_params['cuda']:
        logger.info('Loading a GPU-trained model on GPU')
        checkpoint = torch.load(checkpoint_filepath)
    elif torch.cuda.is_available() and not model_params['cuda']:
        logger.info('Loading a CPU-trained model on GPU')
        checkpoint = torch.load(checkpoint_filepath,
                                map_location='cuda:0')
    else:
        logger.info('Loading a CPU-trained model on CPU')
        checkpoint = torch.load(checkpoint_filepath)
    encoder = Encoder(model_type=checkpoint['encoder']['model_type'],
                      input_size=checkpoint['encoder']['input_size'],
                      hidden_size=checkpoint['encoder']['hidden_size'],
                      num_layers=checkpoint['encoder']['num_layers'],
                      nonlinearity=checkpoint['encoder']['nonlinearity'],
                      bias=checkpoint['encoder']['bias'],
                      dropout=checkpoint['encoder']['dropout'],
                      bidirectional=checkpoint['encoder']['bidirectional'])
    if checkpoint['with_attention']:
        decoder = Attention(model_type=checkpoint['decoder']['model_type'],
                            hidden_size=checkpoint['decoder']['hidden_size'],
                            output_size=checkpoint['decoder']['output_size'],
                            max_seq_len=dataset_params['max_seq_len'],
                            num_layers=checkpoint['decoder']['num_layers'],
                            nonlinearity=checkpoint['decoder']['nonlinearity'],
                            bias=checkpoint['decoder']['bias'],
                            dropout=checkpoint['decoder']['dropout'],
                            bidirectional=checkpoint['decoder']['bidirectional'])
    else:
        decoder = Decoder(model_type=checkpoint['decoder']['model_type'],
                          hidden_size=checkpoint['decoder']['hidden_size'],
                          output_size=checkpoint['decoder']['output_size'],
                          num_layers=checkpoint['decoder']['num_layers'],
                          nonlinearity=checkpoint['decoder']['nonlinearity'],
                          bias=checkpoint['decoder']['bias'],
                          dropout=checkpoint['decoder']['dropout'],
                          bidirectional=checkpoint['decoder']['bidirectional'])
    encoder.load_state_dict(checkpoint['encoder_state_dict'])
    decoder.load_state_dict(checkpoint['decoder_state_dict'])
    if torch.cuda.is_available():
        encoder.to(const.DEVICE)
        decoder.to(const.DEVICE)
    encoder.eval()
    decoder.eval()
    indexes = putils.index_dataset(
        args.data, source_vocab.item2idx, target_vocab.item2idx,
        dataset_params['is_character_based'], dataset_params['max_seq_len'],
        dataset_params['is_reversed'])
    if args.random > 0:
        random.shuffle(indexes)
        for seq_num in range(args.random):
            seq = indexes[seq_num]
            print('-'*80)
            print('>', ' '.join([source_vocab.idx2item[idx]
                                 for idx in seq[0]]))
            print('=', ' '.join([target_vocab.idx2item[idx]
                                 for idx in seq[1]]))
            # TODO: add support for OOV
            predicted_idx, _ = _decode(seq[0], encoder, decoder,
                                       checkpoint['with_attention'],
                                       dataset_params['max_seq_len'])
            print('<', ' '.join([target_vocab.idx2item[idx]
                                 for idx in predicted_idx]))
    else:
        _evaluate(indexes, encoder, decoder, target_vocab, checkpoint,
                  dataset_params)


def main():
    """Launch ortografix."""
    parser = argparse.ArgumentParser(prog='ortografix')
    subparsers = parser.add_subparsers()
    parser_train = subparsers.add_parser(
        'train', formatter_class=argparse.RawTextHelpFormatter,
        help='train the seq2seq model')
    parser_train.set_defaults(func=train)
    parser_train.add_argument('-d', '--data', required=True,
                              help='absolute path to training data')
    parser_train.add_argument('-t', '--model-type',
                              choices=['rnn', 'gru', 'lstm'],
                              default='gru',
                              help='encoder/decoder model type')
    parser_train.add_argument('-c', '--character-based', action='store_true',
                              help='if set, will switch from token-based to '
                                   'character-based model. To be used only '
                                   'for ortographic simplification, not '
                                   'neural machine translation')
    parser_train.add_argument('-s', '--shuffle', action='store_true',
                              help='if set, will shuffle the training data')
    parser_train.add_argument('-m', '--max-seq-len', type=int, default=0,
                              help='maximum sequence length to retain. If not '
                                   'set manually, will be set to the length '
                                   'of the longest sequence in the dataset')
    parser_train.add_argument('-z', '--hidden-size', type=int, default=256,
                              help='size of the hidden layer')
    parser_train.add_argument('-n', '--num-layers', type=int, default=1,
                              help='number of layers to stack in the '
                                   'encoder/decoder models')
    parser_train.add_argument('-l', '--nonlinearity', choices=['tanh', 'relu'],
                              default='tanh', help='activation function to '
                                                   'use. For RNN model only')
    parser_train.add_argument('-b', '--bias', action='store_true',
                              help='whether or not to use biases in '
                                   'encoder/decoder models')
    parser_train.add_argument('-o', '--dropout', type=float, default=0,
                              help='probability in Dropout layer')
    parser_train.add_argument('-i', '--bidirectional', action='store_true',
                              help='if set, will use a bidirectional model '
                                   'in both encoder and decoder models')
    parser_train.add_argument('-r', '--learning-rate', type=float,
                              default=0.01, help='learning rate')
    parser_train.add_argument('-e', '--epochs', type=int, default=1,
                              help='number of epochs')
    parser_train.add_argument('-p', '--print-every', type=int, default=1000,
                              help='how often to print out loss information')
    parser_train.add_argument('-u', '--use-teacher-forcing',
                              action='store_true',
                              help='if set, will use teacher forcing')
    parser_train.add_argument('-f', '--teacher-forcing-ratio', type=float,
                              default=0.5, help='teacher forcing ratio')
    parser_train.add_argument('-a', '--output-dirpath', required=True,
                              help='absolute dirpath where to save models')
    parser_train.add_argument('-w', '--with-attention', action='store_true',
                              help='if set, will use attention-based decoding')
    parser_train.add_argument('-v', '--reverse', action='store_true',
                              help='if set, will reverse dataset pairs')
    parser_evaluate = subparsers.add_parser(
        'evaluate', formatter_class=argparse.RawTextHelpFormatter,
        help='evaluate model on input test set')
    parser_evaluate.set_defaults(func=evaluate)
    parser_evaluate.add_argument('-m', '--model', required=True,
                                 help='absolute path to model directory')
    parser_evaluate.add_argument('-d', '--data', required=True,
                                 help='absolute path to test set')
    parser_evaluate.add_argument('-v', '--reverse', action='store_true',
                                 help='if set, will reverse dataset pairs')
    parser_evaluate.add_argument('-r', '--random', type=int, default=0,
                                 help='if > 0, will test on n sequences '
                                      'randomly selected from test set')
    args = parser.parse_args()
    args.func(args)



#!/usr/bin/env python3
"""ortografix setup.py.

This file details modalities for packaging the ortografix application.
"""

from setuptools import setup

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='ortografix',
    description='Seq2seq model with attention for automatic orthographic simplification',
    author=' Alexandre Kabbach',
    author_email='akb@3azouz.net',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='0.3.4',
    url='https://github.com/akb89/ortografix',
    download_url='https://github.com/akb89/ortografix',
    license='MIT',
    keywords=['seq2seq', 'ortographic simplification'],
    platforms=['any'],
    packages=['ortografix', 'ortografix.logging', 'ortografix.exceptions',
              'ortografix.utils', 'ortografix.model'],
    package_data={'ortografix': ['logging/*.yml', 'resources/*']},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'ortografix = ortografix.main:main'
        ],
    },
    install_requires=['pyyaml>=4.2b1', 'torch==1.4.0', 'textdistance==4.2.0'],
    classifiers=['Development Status :: 2 - Pre-Alpha',
                 'Environment :: Web Environment',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Education',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: MIT License',
                 'Natural Language :: English',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3.8',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence',
                 'Topic :: Software Development :: Libraries :: Python Modules'],
    zip_safe=False,
)



"""Encoder."""
import logging

import torch

import ortografix.utils.constants as const

__all__ = ('Encoder')

logger = logging.getLogger(__name__)


# pylint: disable=R0902
class Encoder(torch.nn.Module):
    """Encoder class."""

    def __init__(self, model_type, input_size, hidden_size, num_layers=1,
                 nonlinearity='tanh', bias=True, dropout=0,
                 bidirectional=False):
        """Initialize encoder model."""
        if model_type not in ['rnn', 'gru', 'lstm']:
            raise Exception('Unsupported model type: {}'.format(model_type))
        super(Encoder, self).__init__()
        self.model_type = model_type
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.nonlinearity = nonlinearity
        self.bias = bias
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.embedding = torch.nn.Embedding(input_size, hidden_size)
        if self.model_type == 'rnn':
            self.rnn = torch.nn.RNN(
                input_size=hidden_size, hidden_size=hidden_size,
                num_layers=num_layers, nonlinearity=nonlinearity,
                bias=bias, batch_first=False, dropout=dropout,
                bidirectional=bidirectional)
        if self.model_type == 'gru':
            self.gru = torch.nn.GRU(
                input_size=hidden_size, hidden_size=hidden_size,
                num_layers=num_layers, bias=bias, batch_first=False,
                dropout=dropout, bidirectional=bidirectional)
        if self.model_type == 'lstm':
            self.lstm = torch.nn.LSTM(
                input_size=hidden_size, hidden_size=hidden_size,
                num_layers=num_layers, bias=bias, batch_first=False,
                dropout=dropout, bidirectional=bidirectional)

    # pylint: disable=R1710, W0221
    def forward(self, input_tensor, hidden_tensor):
        """Apply forward computation."""
        embedded_tensor = self.embedding(input_tensor).view(1, 1, -1)
        if self.model_type == 'rnn':
            return self.rnn(embedded_tensor, hidden_tensor)
        if self.model_type == 'gru':
            return self.gru(embedded_tensor, hidden_tensor)
        if self.model_type == 'lstm':
            return self.lstm(embedded_tensor, hidden_tensor)

    def init_hidden(self):
        """Initialize hidden layers."""
        if self.bidirectional:
            return torch.zeros(self.num_layers*2, 1, self.hidden_size,
                               device=const.DEVICE)
        return torch.zeros(self.num_layers, 1, self.hidden_size,
                           device=const.DEVICE)
